# tests/test_hyphenation.py
import pytest
from etpgrf import Hyphenator
from etpgrf.config import CHAR_SHY

# --- Тестовые данные для русского языка ---
# Формат: (входное_слово, ожидаемый_результат_с_переносами)
# Используем {CHAR_SHY} - это Unicode-представление мягкого переноса (&shy;)
RUSSIAN_HYPHENATION_CASES = [
    ("дом", "дом"),  # Сочень короткое (короче max_unhyphenated_len) не должно меняться
    ("проверка", f"про{CHAR_SHY}верка"),
    ("тестирование", f"тести{CHAR_SHY}рова{CHAR_SHY}ние"),
    ("благотворительностью", f"бла{CHAR_SHY}готво{CHAR_SHY}ритель{CHAR_SHY}ностью"), # Слово с переносом на мягкий знак
    ("фотоаппаратура", f"фотоап{CHAR_SHY}пара{CHAR_SHY}тура"), # проверка слова со сдвоенной согласной
    ("программирование", f"про{CHAR_SHY}грам{CHAR_SHY}миро{CHAR_SHY}вание"),  # слова со сдвоенной согласной
    ("сверхзвуковой", f"сверх{CHAR_SHY}зву{CHAR_SHY}ковой"),
    ("автомобиль", f"авто{CHAR_SHY}мобиль"),
    ("интернационализация", f"инте{CHAR_SHY}рнаци{CHAR_SHY}онали{CHAR_SHY}зация"),
    ("электронный", f"элек{CHAR_SHY}трон{CHAR_SHY}ный"),
    ("информационный", f"инфо{CHAR_SHY}рма{CHAR_SHY}цион{CHAR_SHY}ный"),
    ("автоматизация", f"автома{CHAR_SHY}тиза{CHAR_SHY}ция"),
    ("многоклеточный", f"мно{CHAR_SHY}гокле{CHAR_SHY}точный"),
    ("многофункциональный", f"мно{CHAR_SHY}гофун{CHAR_SHY}кцио{CHAR_SHY}наль{CHAR_SHY}ный"),
    ("непрерывность", f"непре{CHAR_SHY}рывно{CHAR_SHY}сть"),
    ("сверхпроводимость", f"сверх{CHAR_SHY}прово{CHAR_SHY}димо{CHAR_SHY}сть"),
    ("многообразие", f"мно{CHAR_SHY}гоо{CHAR_SHY}бра{CHAR_SHY}зие"),
    ("противоречивость", f"про{CHAR_SHY}тиво{CHAR_SHY}речи{CHAR_SHY}вость"),
    ("непревзойденный", f"непре{CHAR_SHY}взой{CHAR_SHY}ден{CHAR_SHY}ный"),
    ("многослойный", f"мно{CHAR_SHY}гослой{CHAR_SHY}ный"),
    ("суперкомпьютер", f"супе{CHAR_SHY}рко{CHAR_SHY}мпью{CHAR_SHY}тер"), # Неправильный перенос (нужен словарь "приставок/корней/суффиксов")
    ("сверхчувствительный", f"свер{CHAR_SHY}хчув{CHAR_SHY}стви{CHAR_SHY}тель{CHAR_SHY}ный"),  # Неправильный перенос
    ("гиперподъездной", f"гипе{CHAR_SHY}рпо{CHAR_SHY}дъез{CHAR_SHY}дной"),  # Неправильный перенос
]


@pytest.mark.parametrize("input_word, expected_output", RUSSIAN_HYPHENATION_CASES)
def test_russian_word_hyphenation(input_word, expected_output):
    """
    Проверяет ПОВЕДЕНИЕ: правильная расстановка переносов в отдельных русских словах.
    """
    # Arrange (подготовка)
    hyphenator_ru = Hyphenator(langs='ru', max_unhyphenated_len=5, min_tail_len=3)
    # Act (действие) - тестируем самый "атомарный" метод
    actual_output = hyphenator_ru.hyp_in_word(input_word)
    # Assert (проверка)
    assert actual_output == expected_output


ENGLISH_HYPHENATION_CASES = [
    ("color", "color"),  # Короткое слово, не должно меняться
    ("throughout", "throughout"),  # Длинное слово, но из-за икс-графа "ough" не будет переноситься
    ("ambrella", f"amb{CHAR_SHY}rella"),
    ("unbelievable", f"unbel{CHAR_SHY}iev{CHAR_SHY}able"),  # Проверка переноса перед суффиксом "able"
    ("acknowledgment", f"ack{CHAR_SHY}now{CHAR_SHY}ledg{CHAR_SHY}ment"),  # Проверка переноса перед суффиксом "ment"
    ("friendship", f"frien{CHAR_SHY}dship"),  # Проверка переноса перед суффиксом "ship"
    ("thoughtful", f"though{CHAR_SHY}tful"),  #
    ("psychology", f"psy{CHAR_SHY}cho{CHAR_SHY}logy"),  # Проверка переноса после "psy"
    ("extraordinary", f"ext{CHAR_SHY}raor{CHAR_SHY}din{CHAR_SHY}ary"),  # Проверка сложного слова
    ("unbreakable", f"unb{CHAR_SHY}rea{CHAR_SHY}kable"),  # Проверка переноса перед "able"
    ("acknowledgement", f"ack{CHAR_SHY}now{CHAR_SHY}ledge{CHAR_SHY}ment"),  # Проверка икс-графа "dge"
    ("misunderstanding", f"mis{CHAR_SHY}under{CHAR_SHY}stan{CHAR_SHY}ding"),  # Проверка сложного слова
    ("floccinaucinihilipilification", f"floc{CHAR_SHY}cin{CHAR_SHY}auc{CHAR_SHY}inih{CHAR_SHY}ili{CHAR_SHY}pili{CHAR_SHY}fica{CHAR_SHY}tion"),
]

@pytest.mark.parametrize("input_word, expected_output", ENGLISH_HYPHENATION_CASES)
def test_english_word_hyphenation(input_word, expected_output):
    """
    Проверяет ПОВЕДЕНИЕ: правильная расстановка переносов в отдельных английских словах.
    """
    # Arrange (подготовка)
    hyphenator_en = Hyphenator(langs='en', max_unhyphenated_len=5, min_tail_len=3)
    # Act (действие) - тестируем самый "атомарный" метод
    actual_output = hyphenator_en.hyp_in_word(input_word)
    # Assert (проверка)
    assert actual_output == expected_output




# tests/test_hanging.py
# Тесты для модуля висячей пунктуации (HangingPunctuationProcessor).

import pytest
from bs4 import BeautifulSoup
from etpgrf.hanging import HangingPunctuationProcessor
from etpgrf.config import (
    CHAR_RU_QUOT1_OPEN, CHAR_RU_QUOT1_CLOSE,
    CHAR_EN_QUOT1_OPEN, CHAR_EN_QUOT1_CLOSE
)

# Вспомогательная функция для создания soup
def make_soup(html_str):
    return BeautifulSoup(html_str, 'html.parser')

# Набор тестовых случаев в формате:
# (режим, входной_html, ожидаемый_html)
HANGING_TEST_CASES = [
    # --- Режим 'left' (только левая пунктуация) ---
    ('left', f'<p>{CHAR_RU_QUOT1_OPEN}Цитата{CHAR_RU_QUOT1_CLOSE}</p>',
             f'<p><span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>Цитата{CHAR_RU_QUOT1_CLOSE}</p>'),
    ('left', f'<p>(Скобки)</p>',
             f'<p><span class="etp-lpar">(</span>Скобки)</p>'),
    # Правая пунктуация игнорируется
    ('left', f'<p>Текст.</p>', f'<p>Текст.</p>'),

    # --- Режим 'right' (только правая пунктуация) ---
    ('right', f'<p>{CHAR_RU_QUOT1_OPEN}Цитата{CHAR_RU_QUOT1_CLOSE}</p>',
              f'<p>{CHAR_RU_QUOT1_OPEN}Цитата<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></p>'),
    ('right', f'<p>Текст.</p>',
              f'<p>Текст<span class="etp-r-dot">.</span></p>'),
    # Левая пунктуация игнорируется
    ('right', f'<p>(Скобки)</p>', f'<p>(Скобки<span class="etp-rpar">)</span></p>'),

    # --- Режим 'both' (и левая, и правая) ---
    ('both', f'<p>{CHAR_RU_QUOT1_OPEN}Цитата{CHAR_RU_QUOT1_CLOSE}</p>',
             f'<p><span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>Цитата<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></p>'),
    ('both', f'<p>Текст.</p>',
             f'<p>Текст<span class="etp-r-dot">.</span></p>'),
    # Последовательность символов (точка + кавычка)
    ('both', f'<p>Текст.{CHAR_RU_QUOT1_CLOSE}</p>',
             f'<p>Текст<span class="etp-r-dot">.</span><span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></p>'),
    # Вложенные теги
    ('both', f'<p><b>{CHAR_RU_QUOT1_OPEN}Жирный{CHAR_RU_QUOT1_CLOSE}</b></p>',
             f'<p><b><span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>Жирный<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></b></p>'),
    # Смешанный контент
    ('both', f'<p>{CHAR_RU_QUOT1_OPEN}Начало <i>курсив</i> конец.{CHAR_RU_QUOT1_CLOSE}</p>',
             f'<p><span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>Начало <i>курсив</i> конец<span class="etp-r-dot">.</span><span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></p>'),

    # --- Режим None / False (отключено) ---
    (None, f'<p>{CHAR_RU_QUOT1_OPEN}Текст{CHAR_RU_QUOT1_CLOSE}</p>',
           f'<p>{CHAR_RU_QUOT1_OPEN}Текст{CHAR_RU_QUOT1_CLOSE}</p>'),
    (False, f'<p>{CHAR_RU_QUOT1_OPEN}Текст{CHAR_RU_QUOT1_CLOSE}</p>',
            f'<p>{CHAR_RU_QUOT1_OPEN}Текст{CHAR_RU_QUOT1_CLOSE}</p>'),

    # --- Отсутствие висячих символов ---
    ('both', '<p>Простой текст без спецсимволов!</p>', '<p>Простой текст без спецсимволов!</p>'),

    # --- Проверка контекста (пробелы) ---
    # 1. Левая кавычка внутри слова (не должна висеть)
    ('both', f'<p>func{CHAR_RU_QUOT1_OPEN}arg{CHAR_RU_QUOT1_CLOSE}</p>',
             f'<p>func{CHAR_RU_QUOT1_OPEN}arg<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></p>'), # Правая висит, т.к. конец узла
    # 2. Правая кавычка внутри слова (не должна висеть)
    ('both', f'<p>1{CHAR_RU_QUOT1_CLOSE}2</p>',
             f'<p>1{CHAR_RU_QUOT1_CLOSE}2</p>'),
    # 3. Левая кавычка после пробела (должна висеть)
    ('both', f'<p>func {CHAR_RU_QUOT1_OPEN}arg</p>',
             f'<p>func <span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>arg</p>'),
    # 4. Правая кавычка перед пробелом (должна висеть)
    ('both', f'<p>arg{CHAR_RU_QUOT1_CLOSE} next</p>',
             f'<p>arg<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span> next</p>'),
    # 5. Точка внутри числа (не должна висеть)
    ('both', '<p>3.14</p>', '<p>3.14</p>'),
    # 6. Точка в конце предложения (должна висеть)
    ('both', '<p>End.</p>', '<p>End<span class="etp-r-dot">.</span></p>'),
]


@pytest.mark.parametrize("mode, input_html, expected_html", HANGING_TEST_CASES)
def test_hanging_punctuation_processor(mode, input_html, expected_html):
    """
    Проверяет работу HangingPunctuationProcessor в различных режимах.
    """
    # Arrange
    processor = HangingPunctuationProcessor(mode=mode)
    soup = make_soup(input_html)

    # Act
    processor.process(soup)
    actual_html = str(soup)

    # Assert
    assert actual_html == expected_html


def test_hanging_punctuation_target_tags():
    """
    Отдельный тест для проверки работы со списком целевых тегов.
    """
    mode = ['blockquote', 'h1']
    input_html = (f'<div>{CHAR_RU_QUOT1_OPEN}Игнор{CHAR_RU_QUOT1_CLOSE}</div>'
                  f'<blockquote>{CHAR_RU_QUOT1_OPEN}Обработка{CHAR_RU_QUOT1_CLOSE}</blockquote>'
                  f'<h1>{CHAR_RU_QUOT1_OPEN}Заголовок{CHAR_RU_QUOT1_CLOSE}</h1>')
    
    expected_html = (f'<div>{CHAR_RU_QUOT1_OPEN}Игнор{CHAR_RU_QUOT1_CLOSE}</div>'
                     f'<blockquote><span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>Обработка<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></blockquote>'
                     f'<h1><span class="etp-laquo">{CHAR_RU_QUOT1_OPEN}</span>Заголовок<span class="etp-raquo">{CHAR_RU_QUOT1_CLOSE}</span></h1>')

    processor = HangingPunctuationProcessor(mode=mode)
    soup = make_soup(input_html)
    
    processor.process(soup)
    
    assert str(soup) == expected_html



# etpgrf/unbreakables.py
# Модуль для предотвращения "висячих" предлогов, союзов и других коротких слов в начале строки.
# Он "приклеивает" такие слова к последующему слову с помощью неразрывного пробела.
# Кстати в русском тексте союзы составляют 7,61%


import regex
import logging
import html
from etpgrf.config import LANG_RU, LANG_RU_OLD, LANG_EN  # , KEY_NBSP, ALL_ENTITIES
from etpgrf.comutil import parse_and_validate_langs
from etpgrf.config import CHAR_NBSP
from etpgrf.defaults import etpgrf_settings

# --- Наборы коротких слов для разных языков ---
# Используем frozenset для скорости и неизменяемости.
# Слова в нижнем регистре для удобства сравнения.

_RU_UNBREAKABLE_WORDS = frozenset([
    # Предлоги (только короткие... длинные, типа `ввиду`, `ввиду` и т.п., могут быть "висячими")
    'в', 'без', 'до', 'из', 'к', 'на', 'по', 'о', 'от', 'перед', 'при', 'через', 'с', 'у', 'за', 'над',
    'об', 'под', 'про', 'для', 'ко', 'со', 'без', 'то', 'во', 'из-за', 'из-под', 'как',
    # Союзы (без сложных, тип `как будто`, `как если бы`, `за то` и т.п.)
    'и', 'а', 'но', 'да',
    # Частицы
    'не', 'ни',
    # Местоимения
    'я', 'ты', 'он', 'мы', 'вы', 'им', 'их', 'ей', 'ею',
    # Устаревшие или специфичные
    'сей', 'сия', 'сие',
])

# Постпозитивные частицы, которые приклеиваются к ПРЕДЫДУЩЕМУ слову
_RU_POSTPOSITIVE_PARTICLES = frozenset([
    'ли', 'ль', 'же', 'ж', 'бы', 'б',
])

# Для дореформенной орфографии можно добавить специфичные слова, если нужно
_RU_OLD_UNBREAKABLE_WORDS = _RU_UNBREAKABLE_WORDS | frozenset([
    'і', 'безъ', 'черезъ', 'въ', 'изъ', 'къ', 'отъ', 'съ', 'надъ', 'подъ', 'объ', 'какъ',
    'сiя', 'сiе', 'сiй', 'онъ', 'тъ',
])

# Постпозитивные частицы, которые приклеиваются к ПРЕДЫДУЩЕМУ слову
_RU_OLD_POSTPOSITIVE_PARTICLES = frozenset([
    'жъ', 'бъ'
])

_EN_UNBREAKABLE_WORDS = frozenset([
    # 1-2 letter words (I - as pronoun)
    'a', 'an', 'as', 'at', 'by', 'in', 'is', 'it', 'of', 'on', 'or', 'so', 'to', 'if',
    # 3-4 letter words
    'for', 'from', 'into', 'that', 'then', 'they', 'this', 'was', 'were', 'what', 'when', 'with',
    'not', 'but', 'which', 'the'
])

# --- Настройки логирования ---
logger = logging.getLogger(__name__)


# --- Класс Unbreakables (обработка неразрывных конструкций) ---
class Unbreakables:
    """
    Правила обработки коротких слов (предлогов, союзов, частиц и местоимений) для предотвращения их отрыва
    от последующих слов.
    """

    def __init__(self, langs: str | list[str] | tuple[str, ...] | frozenset[str] | None = None):
        self.langs = parse_and_validate_langs(langs)

        # --- 1. Собираем наборы слов для обработки ---
        pre_words = set()
        post_words = set()
        # Собираем слова которые должны быть приклеены
        if LANG_RU in self.langs:
            pre_words.update(_RU_UNBREAKABLE_WORDS)
            post_words.update(_RU_POSTPOSITIVE_PARTICLES)
        if LANG_RU_OLD in self.langs:
            pre_words.update(_RU_OLD_UNBREAKABLE_WORDS)
            post_words.update(_RU_OLD_POSTPOSITIVE_PARTICLES)
        if LANG_EN in self.langs:
            pre_words.update(_EN_UNBREAKABLE_WORDS)

        # Собираем единый набор слов с пост-позиционными словами (не отрываются от предыдущих слов)
        # Убедимся, что пост-позиционные слова не обрабатываются дважды
        pre_words -= post_words

        # --- 2. Компиляция паттернов с оптимизацией ---
        self._pre_pattern = None
        if pre_words:
            # Оптимизация: сортируем слова по длине от большего к меньшему
            sorted_words = sorted(list(pre_words), key=len, reverse=True)
            # Паттерн для слов, ПОСЛЕ которых нужен nbsp. regex.escape для безопасности.
            self._pre_pattern = regex.compile(r"(?i)\b(" + "|".join(map(regex.escape, sorted_words)) + r")\b\s+")

        self._post_pattern = None
        if post_words:
            # Оптимизация: сортируем слова по длине от большего к меньшему
            sorted_particles = sorted(list(post_words), key=len, reverse=True)
            # Паттерн для слов, ПЕРЕД которыми нужен nbsp.
            self._post_pattern = regex.compile(r"(?i)(\s)\b(" + "|".join(map(regex.escape, sorted_particles)) + r")\b")

        logger.debug(f"Unbreakables `__init__`. Langs: {self.langs}, "
                      f"Pre-words: {len(pre_words)}, Post-words: {len(post_words)}")


    def process(self, text: str) -> str:
        """
        Заменяет обычные пробелы вокруг коротких слов на неразрывные.
        """
        if not text:
            return text
        processed_text = text

        # 1. Обработка слов, ПОСЛЕ которых нужен неразрывный пробел ("в дом" -> "в&nbsp;дом")
        if self._pre_pattern:
            processed_text = self._pre_pattern.sub(r"\g<1>" + CHAR_NBSP, processed_text)

        # 2. Обработка частиц, ПЕРЕД которыми нужен неразрывный пробел ("сказал бы" -> "сказал&nbsp;бы")
        if self._post_pattern:
            # \g<1> - это пробел, \g<2> - это частица
            processed_text = self._post_pattern.sub(CHAR_NBSP + r"\g<2>", processed_text)

        return processed_text



# tests/test_unbreakables.py
import pytest
from etpgrf import Unbreakables
from etpgrf.config import CHAR_NBSP, CHAR_THIN_SP


# Список русских слов, которые должны "приклеиваться" к следующему слову.
RUSSIAN_PREPOSITIONS_TO_TEST = [
    # Предлоги (только короткие... длинные, типа `ввиду`, `ввиду` и т.п., могут быть "висячими")
    'в', 'без', 'до', 'из', 'к', 'на', 'по', 'о', 'от', 'перед', 'при', 'через', 'с', 'у', 'за', 'над',
    'об', 'под', 'про', 'для', 'ко', 'со', 'без', 'то', 'во', 'из-за', 'из-под', 'как',
    # Союзы (без сложных, тип `как будто`, `как если бы`, `за то` и т.п.)
    'и', 'а', 'но', 'да', 'как',
    # Частицы
    'не', 'ни',
    # Местоимения
    'я', 'ты', 'он', 'мы', 'вы', 'им', 'их', 'ей', 'ею',
    # Устаревшие или специфичные
    'сей', 'сия', 'сие',
]

@pytest.mark.parametrize("word", RUSSIAN_PREPOSITIONS_TO_TEST)
def test_russian_prepositions_are_unbreakable(word):
    """
    Проверяет ПОВЕДЕНИЕ: короткие слова "приклеиваются" к следующему. Для русского языка. Параметризованный тест
    """
    # Arrange (подготовка)
    unbreakables_ru = Unbreakables(langs='ru')
    input_text = f"Проверка {word} тестирование."
    expected_text = f"Проверка {word}{CHAR_NBSP}тестирование."
    # Act (действие, которое выполняем)
    actual_text = unbreakables_ru.process(input_text)
    # Assert (утверждение, что результат соответствует ожиданиям)
    assert actual_text == expected_text


# Список английских слов, которые должны "приклеиваться" к следующему слову.
ENGLISH_PREPOSITIONS_TO_TEST = [
    'a', 'an', 'as', 'at', 'by', 'in', 'is', 'it', 'of', 'on', 'or', 'so', 'to', 'if',
    'for', 'from', 'into', 'that', 'then', 'they', 'this', 'was', 'were', 'what', 'when', 'with',
    'not', 'but', 'which', 'the'
]

@pytest.mark.parametrize("word", ENGLISH_PREPOSITIONS_TO_TEST)
def test_english_prepositions_are_unbreakable(word):
    """
    Проверяет ПОВЕДЕНИЕ: короткие слова "приклеиваются" к следующему. Для английского языка. Параметризованный тест
    """
    # Arrange (подготовка)
    unbreakables_en = Unbreakables(langs='en')
    input_text = f"Training {word} test."
    expected_text = f"Training {word}{CHAR_NBSP}test."
    # Act (действие, которое выполняем)
    actual_text = unbreakables_en.process(input_text)
    # Assert (утверждение, что результат соответствует ожиданиям)
    assert actual_text == expected_text


# Смешанный тест для русского и английского языков
def test_mix_prepositions_are_unbreakable():
    """
    Проверяет ПОВЕДЕНИЕ: короткие слова "приклеиваются" к следующему. Для смешанного русско-английского текста.
    """
    # Arrange (подготовка)
    unbreakables_mix = Unbreakables(langs='ru+en')
    input_text = f"Для the Guardian он написал блестящую статью."
    expected_text = f"Для{CHAR_NBSP}the{CHAR_NBSP}Guardian он{CHAR_NBSP}написал блестящую статью."
    # Act (действие, которое выполняем)
    actual_text = unbreakables_mix.process(input_text)
    # Assert (утверждение, что результат соответствует ожиданиям)
    assert actual_text == expected_text


# Список русских постпозитивных частиц, которые должны "приклеиваться" к предыдущему слову.
RUSSIAN_POSTPOSITIVE_PARTICLES_TO_TEST = [
    'ли', 'ль', 'же', 'ж', 'бы', 'б'
]

@pytest.mark.parametrize("word", RUSSIAN_POSTPOSITIVE_PARTICLES_TO_TEST)
def test_russian_postpositive_particle(word):
    """
    Проверяет ПОВЕДЕНИЕ: русские постпозитивные частицы "приклеиваются" к предыдущему слову.
    """
    # Arrange
    unbreakables_ru = Unbreakables(langs='ru')
    input_text = f"Отчего {word} не поспать?"
    expected_text = f"Отчего{CHAR_NBSP}{word} не{CHAR_NBSP}поспать?"
    # Act
    actual_text = unbreakables_ru.process(input_text)
    # Assert
    assert actual_text == expected_text


# Тесты для проверки особых случаев в Unbreakables
UNBREAKABLES_SPECIAL_TEST_CASES = [
    ('ru', "до н.э.", f"до{CHAR_NBSP}н.э."),
    ('ru', "слово и тогда", f"слово и{CHAR_NBSP}тогда"),
    ('ru', "слово а тогда", f"слово а{CHAR_NBSP}тогда"),
    ('ru', "Проверка и тестирование.", f"Проверка и{CHAR_NBSP}тестирование."),

]


@pytest.mark.parametrize("lang, input_string, expected_output", UNBREAKABLES_SPECIAL_TEST_CASES)
def test_layout_processor_with_options(lang, input_string, expected_output):
    """Проверяет работу Unbreakables с особыми случаями. """
    processor = Unbreakables(langs=lang)
    actual_output = processor.process(input_string)
    assert actual_output == expected_output



# etpgrf/hanging.py
# Модуль для расстановки висячей пунктуации.

import logging
from bs4 import BeautifulSoup, NavigableString, Tag
from .config import (
    HANGING_PUNCTUATION_LEFT_CHARS,
    HANGING_PUNCTUATION_RIGHT_CHARS,
    HANGING_PUNCTUATION_CLASSES
)

logger = logging.getLogger(__name__)


class HangingPunctuationProcessor:
    """
    Оборачивает символы висячей пунктуации в специальные теги <span> с классами.
    """

    def __init__(self, mode: str | bool | list[str] | None = None):
        """
        :param mode: Режим работы:
                     - None / False: отключено.
                     - 'left': только левая пунктуация.
                     - 'right': только правая пунктуация.
                     - 'both' / True: и левая, и правая.
                     - list[str]: список тегов (например, ['p', 'blockquote']),
                       внутри которых применять 'both'.
        """
        self.mode = mode
        self.target_tags = None
        self.active_chars = set()

        # Определяем, какие символы будем обрабатывать
        if isinstance(mode, list):
            self.target_tags = set(t.lower() for t in mode)
            # Если передан список тегов, включаем полный режим ('both') внутри них
            self.active_chars.update(HANGING_PUNCTUATION_LEFT_CHARS)
            self.active_chars.update(HANGING_PUNCTUATION_RIGHT_CHARS)
        elif mode == 'left':
            self.active_chars.update(HANGING_PUNCTUATION_LEFT_CHARS)
        elif mode == 'right':
            self.active_chars.update(HANGING_PUNCTUATION_RIGHT_CHARS)
        elif mode == 'both' or mode is True:
            self.active_chars.update(HANGING_PUNCTUATION_LEFT_CHARS)
            self.active_chars.update(HANGING_PUNCTUATION_RIGHT_CHARS)
        
        # Предварительно фильтруем карту классов, оставляя только активные символы
        self.char_to_class = {
            char: cls 
            for char, cls in HANGING_PUNCTUATION_CLASSES.items()
            if char in self.active_chars
        }

        logger.debug(f"HangingPunctuationProcessor initialized. Mode: {mode}, Active chars count: {len(self.active_chars)}")

    def process(self, soup: BeautifulSoup) -> BeautifulSoup:
        """
        Проходит по дереву soup и оборачивает висячие символы в span.
        """
        if not self.active_chars:
            return soup

        # Если задан список целевых тегов, обрабатываем только их содержимое
        if self.target_tags:
            # Находим все теги из списка
            # Используем select для поиска (например: "p, blockquote, h1")
            selector = ", ".join(self.target_tags)
            roots = soup.select(selector)
        else:
            # Иначе обрабатываем весь документ (начиная с корня)
            roots = [soup]

        for root in roots:
            self._process_node_recursive(root, soup)

        return soup

    def _process_node_recursive(self, node, soup):
        """
        Рекурсивно обходит узлы. Если находит NavigableString с нужными символами,
        разбивает его и вставляет span'ы.
        """
        # Работаем с копией списка детей, так как будем менять структуру дерева на лету
        # (replace_with меняет дерево)
        if hasattr(node, 'children'):
            for child in list(node.children):
                if isinstance(child, NavigableString):
                    self._process_text_node(child, soup)
                elif isinstance(child, Tag):
                    # Не заходим внутрь тегов, которые мы сами же и создали (или аналогичных),
                    # чтобы избежать рекурсивного ада, хотя классы у нас специфичные.
                    self._process_node_recursive(child, soup)

    def _process_text_node(self, text_node: NavigableString, soup: BeautifulSoup):
        """
        Анализирует текстовый узел. Если в нем есть символы для висячей пунктуации,
        заменяет узел на фрагмент (список узлов), где эти символы обернуты в span.
        """
        text = str(text_node)
        
        # Быстрая проверка: если в тексте вообще нет ни одного нашего символа, выходим
        if not any(char in text for char in self.active_chars):
            return

        # Если символы есть, нам нужно "разобрать" строку.
        new_nodes = []
        current_text_buffer = ""
        text_len = len(text)

        for i, char in enumerate(text):
            if char in self.char_to_class:
                should_hang = False
                
                # Проверяем контекст (пробелы или другие висячие символы вокруг)
                if char in HANGING_PUNCTUATION_LEFT_CHARS:
                    # Левая пунктуация:
                    # 1. Начало узла
                    # 2. Перед ней пробел
                    # 3. Перед ней другой левый висячий символ (например, "((text")
                    if (i == 0 or 
                        text[i-1].isspace() or 
                        text[i-1] in HANGING_PUNCTUATION_LEFT_CHARS):
                        should_hang = True
                elif char in HANGING_PUNCTUATION_RIGHT_CHARS:
                    # Правая пунктуация:
                    # 1. Конец узла
                    # 2. После нее пробел
                    # 3. После нее другой правый висячий символ (например, "text.»")
                    if (i == text_len - 1 or 
                        text[i+1].isspace() or 
                        text[i+1] in HANGING_PUNCTUATION_RIGHT_CHARS):
                        should_hang = True
                
                if should_hang:
                    # 1. Сбрасываем накопленный буфер текста (если есть)
                    if current_text_buffer:
                        new_nodes.append(NavigableString(current_text_buffer))
                        current_text_buffer = ""
                    
                    # 2. Создаем span для висячего символа
                    span = soup.new_tag("span")
                    span['class'] = self.char_to_class[char]
                    span.string = char
                    new_nodes.append(span)
                else:
                    # Если контекст не подходит, оставляем символ как обычный текст
                    current_text_buffer += char
            else:
                # Просто накапливаем символ
                current_text_buffer += char

        # Добавляем остаток буфера
        if current_text_buffer:
            new_nodes.append(NavigableString(current_text_buffer))

        # Заменяем исходный текстовый узел на набор новых узлов.
        if new_nodes:
            first_node = new_nodes[0]
            text_node.replace_with(first_node)
            
            # Остальные вставляем последовательно после первого
            current_pos = first_node
            for next_node in new_nodes[1:]:
                current_pos.insert_after(next_node)
                current_pos = next_node



# tests/test_quotes.py
# Тесты для модуля QuotesProcessor. Проверяют обработку кавычек в тексте.

import pytest
from etpgrf.quotes import QuotesProcessor

# Набор тестовых случаев в формате:
# (язык, входная_строка, ожидаемый_результат)
QUOTES_TEST_CASES = [
    # --- Базовые случаи ---
    ('ru', 'Текст \"в кавычках\"', 'Текст «в кавычках»'),
    ('en', 'Text \"in quotes\"', 'Text “in quotes”'),
    ('ru', '(\"в скобках\")', '(«в скобках»)'),

    # --- Случаи, которые не должны меняться ---
    ('ru', 'Текст без кавычек.', 'Текст без кавычек.'),
    ('ru', 'Уже есть «ёлочки» и „лапки“', 'Уже есть «ёлочки» и „лапки“'),
    ('en', 'Already have “curly” and ‘single’ quotes', 'Already have “curly” and ‘single’ quotes'),

    # --- Сложные случаи и исключения (QuotesProcessor должен их игнорировать) ---
    ('ru', 'Размер экрана 17"', 'Размер экрана 17"'),
    ('en', 'Screen size 17"', 'Screen size 17"'),
    ('ru', 'Код print("Hello, World!")', 'Код print(«Hello, World!»)'), # Скобки помогают определить контекст
    ('ru', ' " одинокая кавычка " ', ' " одинокая кавычка " '), # Кавычки окружены пробелами
    ('ru', 'слово"вплотную', 'слово"вплотную'),
    ('ru', 'вплотную"слово', 'вплотную"слово'),

    # --- Вложенность и несколько пар ---
    ('ru', 'Он сказал: "Привет, мир!"', 'Он сказал: «Привет, мир!»'),
    ('ru', 'Она ответила: "И тебе "привет"!"', 'Она ответила: «И тебе «привет»!»'),

    # --- Обработка пунктуации ---
    # Точка СНАРУЖИ кавычек - правильная пунктуация, корректно обрабатывается
    ('ru', 'Текст "в кавычках".', 'Текст «в кавычках».'),
    ('en', '"Word".', '“Word”.'),
    # Точка ВНУТРИ кавычек
    ('ru', 'Текст "в кавычках."', 'Текст «в кавычках.»'),
    ('en', '"Word."', '“Word.”'),
    # Знаки ?, !, … ВНУТРИ кавычек - правильная пунктуация, корректно обрабатывается
    ('ru', '"Слово?"', '«Слово?»'),
    ('en', '"Word?"', '“Word?”'),
    ('ru', '"Слово!"', '«Слово!»'),
    ('en', '"Word!"', '“Word!”'),
    ('ru', '"Слово…"', '«Слово…»'),
    ('en', '"Word…"', '“Word…”'),
    # Запятая СНАРУЖИ кавычек
    ('ru', '"Слово", - сказал он.', '«Слово», - сказал он.'),
    # Цифры в кавычках не обрабатываются
    ('ru', 'Цифры "123" в кавычках', 'Цифры "123" в кавычках'),
    ('en', 'Numbers "123" in quotes', 'Numbers "123" in quotes'),

    # --- Языки и приоритеты ---
    (None, 'Текст "в кавычках"', 'Текст «в кавычках»'),  # Язык не указан -> используется DEFAULT_LANGS[0] ('ru')
    ('ru-en', 'Текст "в кавычках"', 'Текст «в кавычках»'),  # 'ru' первый в списке -> используются русские кавычки
    ('en-ru', 'Text "in quotes"', 'Text “in quotes”'),  # 'en' первый в списке -> используются английские кавычки
]


@pytest.mark.parametrize("lang, input_string, expected_output", QUOTES_TEST_CASES)
def test_quotes_processor(lang, input_string, expected_output):
    """
    Проверяет работу QuotesProcessor в изоляции.
    """
    # Arrange (подготовка)
    # Создаем экземпляр процессора только для этого теста
    processor = QuotesProcessor(langs=lang)

    # Act (действие)
    # Выполняем обработку
    actual_output = processor.process(input_string)

    # Assert (проверка)
    # Сравниваем результат с ожидаемым
    assert actual_output == expected_output


# etpgrf/sanitizer.py
# Модуль для очистки и нормализации HTML-кода перед типографикой.

import logging
from bs4 import BeautifulSoup
from .config import (SANITIZE_ALL_HTML, SANITIZE_ETPGRF, SANITIZE_NONE,
                     HANGING_PUNCTUATION_CLASSES, PROTECTED_HTML_TAGS)

logger = logging.getLogger(__name__)


class SanitizerProcessor:
    """
    Выполняет очистку HTML-кода в соответствии с заданным режимом.
    """

    def __init__(self, mode: str | bool | None = SANITIZE_NONE):
        """
        :param mode: Режим очистки:
                     - 'etp' (SANITIZE_ETPGRF): удаляет только разметку висячей пунктуации.
                     - 'html' (SANITIZE_ALL_HTML): удаляет все HTML-теги.
                     - None или False: ничего не делает.
        """
        if mode is False:
            mode = SANITIZE_NONE
        self.mode = mode
        
        # Оптимизация: заранее готовим CSS-селектор для поиска висячей пунктуации
        if self.mode == SANITIZE_ETPGRF:
            # Собираем уникальные классы
            unique_classes = sorted(list(frozenset(HANGING_PUNCTUATION_CLASSES.values())))
            # Формируем селектор вида: span.class1, span.class2, ...
            # Это позволяет использовать нативный парсер (lxml) для поиска, что намного быстрее python-лямбд.
            self._etp_selector = ", ".join(f"span.{cls}" for cls in unique_classes)
        else:
            self._etp_selector = None

        logger.debug(f"SanitizerProcessor `__init__`. Mode: {self.mode}")

    def process(self, soup: BeautifulSoup) -> BeautifulSoup | str:
        """
        Применяет правила очистки к `soup`-объекту.

        :param soup: Объект BeautifulSoup для обработки.
        :return: Обработанный объект BeautifulSoup или строка (в режиме 'html').
        """
        if self.mode == SANITIZE_ETPGRF:
            if not self._etp_selector:
                return soup

            # Используем CSS-селектор для быстрого поиска всех нужных элементов
            spans_to_clean = soup.select(self._etp_selector)

            # "Агрессивная" очистка: просто "разворачиваем" все найденные теги,
            # заменяя их своим содержимым.
            for span in spans_to_clean:
                span.unwrap()

            return soup

        elif self.mode == SANITIZE_ALL_HTML:
            # Оптимизированный подход:
            # 1. Удаляем защищенные теги (script, style и т.д.) вместе с содержимым.
            #    Используем select для поиска, так как это обычно быстрее.
            if PROTECTED_HTML_TAGS:
                # Формируем селектор: script, style, pre, ...
                protected_selector = ", ".join(PROTECTED_HTML_TAGS)
                for tag in soup.select(protected_selector):
                    tag.decompose() # Полное удаление тега из дерева

            # 2. Извлекаем чистый текст из оставшегося дерева.
            #    get_text() работает на уровне C (в lxml) и намного быстрее ручного обхода.
            return soup.get_text()

        # Если режим не задан, ничего не делаем
        return soup



# etpgrf/quotes.py
# Модуль для расстановки кавычек в тексте

import regex
import logging
from .config import (LANG_RU, LANG_EN, CHAR_RU_QUOT1_OPEN, CHAR_RU_QUOT1_CLOSE, CHAR_EN_QUOT1_OPEN,
                     CHAR_EN_QUOT1_CLOSE, CHAR_RU_QUOT2_OPEN, CHAR_RU_QUOT2_CLOSE, CHAR_EN_QUOT2_OPEN,
                     CHAR_EN_QUOT2_CLOSE)
from .comutil import parse_and_validate_langs

# --- Настройки логирования ---
logger = logging.getLogger(__name__)

# Определяем стили кавычек для разных языков
# Формат: (('открывающая_ур1', 'закрывающая_ур1'), ('открывающая_ур2', 'закрывающая_ур2'))
_QUOTE_STYLES = {
    LANG_RU: ((CHAR_RU_QUOT1_OPEN, CHAR_RU_QUOT1_CLOSE), (CHAR_RU_QUOT2_OPEN, CHAR_RU_QUOT2_CLOSE)),
    LANG_EN: ((CHAR_EN_QUOT1_OPEN, CHAR_EN_QUOT1_CLOSE), (CHAR_EN_QUOT2_OPEN, CHAR_EN_QUOT2_CLOSE)),
}


class QuotesProcessor:
    """
    Обрабатывает прямые кавычки ("), превращая их в типографские
    в зависимости от языка и контекста.
    """

    def __init__(self, langs: str | list[str] | tuple[str, ...] | frozenset[str] | None = None):
        self.langs = parse_and_validate_langs(langs)

        # Выбираем стиль кавычек на основе первого поддерживаемого языка
        self.open_quote = '"'
        self.close_quote = '"'

        for lang in self.langs:
            if lang in _QUOTE_STYLES:
                self.open_quote = _QUOTE_STYLES[lang][0][0]
                self.close_quote = _QUOTE_STYLES[lang][0][1]
                logger.debug(
                    f"QuotesProcessor: выбран стиль кавычек для языка '{lang}': '{self.open_quote}...{self.close_quote}'")
                break  # Используем стиль первого найденного языка

        # Паттерн для открывающей кавычки: " перед буквой/цифрой,
        # которой предшествует пробел, начало строки или открывающая скобка.
        # (?<=^|\s|[\(\[„\"‘\']) - "просмотр назад" на начало строки... ищет пробел \s или знак из набора ([„"‘'
        # (?=\p{L})              - "просмотр вперед" на букву \p{L} (но не цифру).
        self._opening_quote_pattern = regex.compile(r'(?<=^|\s|[\(\[„\"‘\'])\"(?=\p{L})')
        # self._opening_quote_pattern = regex.compile(r'(?<=^|\s|\p{Pi}|["\'\(\)])\"(?=\p{L})')

        # Паттерн для закрывающей кавычки: " после буквы/цифры,
        # за которой следует пробел, пунктуация или конец строки.
        # (?<=\p{L}|[?!…\.])        - "просмотр назад" на букву или ?!… и точку.
        # (?=\s|[.,;:!?\)\"»”’]|\Z) - "просмотр вперед" на пробел, пунктуацию или конец строки (\Z).
        self._closing_quote_pattern = regex.compile(r'(?<=\p{L}|[?!…\.])\"(?=\s|[\.,;:!?\)\]»”’\"\']|\Z)')
        # self._closing_quote_pattern = regex.compile(r'(?<=\p{L}|\p{N})\"(?=\s|[\.,;:!?\)\"»”’]|\Z)')
        # self._closing_quote_pattern = regex.compile(r'(?<=\p{L}|[?!…])\"(?=\s|[\p{Po}\p{Pf}"\']|\Z)')

    def process(self, text: str) -> str:
        """
        Применяет правила замены кавычек к тексту.
        """
        if '"' not in text:
            # Быстрый выход, если в тексте нет прямых кавычек
            return text

        processed_text = text

        # 1. Заменяем открывающие кавычки
        # Заменяем только найденную кавычку, так как просмотр вперед не захватывает символы.
        processed_text = self._opening_quote_pattern.sub(self.open_quote, processed_text)

        # 2. Заменяем закрывающие кавычки
        processed_text = self._closing_quote_pattern.sub(self.close_quote, processed_text)

        return processed_text